import java.io.*;
import java.util.ArrayList;

import static java.util.Arrays.sort;

/**
 * Author: Alan P. Sexton, School of Computer Science, University of Birmingham, UK
 * Date: 30/05/12
 * <p />
 * Copyright (c) 2012 by Alan P. Sexton. This work is licensed under the Creative Commons Attribution 3.0 Unported
 * License. To view a copy of this license, visit http://creativecommons.org/licenses/by/3.0/ or send a letter to
 *     Creative Commons,
 *     444 Castro Street,
 *     Suite 900,
 *     Mountain View,
 *     California, 94041,
 *     USA.
 *
 */
public class ClefSegmentEval
{
    public static final String VERSION = "1.0";


    public static void main(String[] args)
    {
        if (args.length < 2)
            usage();
        BufferedReader gtReader = null;
        try
        {
            gtReader = new BufferedReader(new FileReader(args[0]));
        }
        catch (FileNotFoundException e)
        {
            System.err.printf("Error: Cannot open the ground truth input file \"%s\" for reading%n", args[0]);
            System.err.println(e.getMessage());
            usage();
        }

        BufferedReader evalReader = null;
        try
        {
            evalReader = new BufferedReader(new FileReader(args[1]));
        }
        catch (FileNotFoundException e)
        {
            System.err.printf("Error: Cannot open the evaluation set input file \"%s\" for reading%n", args[1]);
            System.err.println(e.getMessage());
            usage();
        }

        int tolerances[] = new int[Math.max(1, args.length - 2)];
        if (args.length < 2)
            tolerances[0] = 0;
        else
        {
            for (int i = 0 ; i < args.length - 2 ; i++)
            {
                int val = 0;
                try
                {
                    val = Integer.parseInt(args[i + 2]);
                    if (val < 0)
                        throw new NumberFormatException("tolerance argument value must not be negative");
                    if (i > 0 && val <= tolerances[i - 1])
                        throw new NumberFormatException("tolerance argument list must be strictly increasing");
                }
                catch (NumberFormatException e)
                {
                    System.err.println("Error in tolerance arguments: " + e.getMessage());
                    usage();
                }
                tolerances[i] = val;
            }
        }

        SegmentBB gtArray[] = SegmentBB.loadAndSort(args[0], gtReader);
        SegmentBB evalArray[] = SegmentBB.loadAndSort(args[1], evalReader);

        System.out.printf("GroundTruthSize,EvalSetSize,Tolerance,TruePositives,FalsePositives,FalseNegatives," +
                          "Precision,Recall,F1Measure,MultipleGTMatches,MultipleEvalMatches%n");
        for (int tolerance : tolerances)
        {
            int matchResults[] = calculateMatches(gtArray, evalArray, tolerance);
            int tp = matchResults[0];
            int fp = matchResults[1];
            int fn = matchResults[2];
            int multipleGTMatches = matchResults[3];
            int multipleEvalMatches = matchResults[4];
            double precision = ((double) tp) / ((double) tp + fp);
            double recall = ((double) tp) / ((double) tp + fn);
            double f1measure = 2d * precision * recall / (precision + recall);
            System.out.printf("%d,%d,%d,%d,%d,%d,%.5f,%.5f,%.5f,%d,%d%n",
                              gtArray.length,
                              evalArray.length,
                              tolerance,
                              tp,
                              fp,
                              fn,
                              precision,
                              recall,
                              f1measure,
                              multipleGTMatches,
                              multipleEvalMatches);
        }


    }

    private static int[] calculateMatches(SegmentBB gtArray[], SegmentBB evalArray[], int tolerance)
    {
        // We want to calculate True Positives, False Positives and False Negatives for matching evaluation segments
        // to ground truth segments
        // We calculate it as follows: Create arrays evalToGt and gtToEval so that
        // evalToGT[i] == j and gtToEval[j] = j if evalArray[i] matches gtArray[j] within the given tolerance. Each
        // such case is a True Positive.
        // evalToGt[i] = -1 if there is no segment in gtArray that matches evArray[i]. This is a False Positive.
        // gtToEval[i] = -1 if there is no segment in evalArray that matches gtArray[i]. This is a False Negative.
        // Then we just have to count the -1 and the non -1 values in the two arrays to get the final results.


        int evalToGt[] = new int[evalArray.length];
        for (int i = 0 ; i < evalToGt.length ; i++)
            evalToGt[i] = -1;

        int gtToEval[] = new int[gtArray.length];
        for (int i = 0 ; i < gtToEval.length ; i++)
            gtToEval[i] = -1;

        int tpEval = 0; // True Positives (matching eval and gt pair) as calculated from evalToGt
        int tpGt = 0;  // True Positives (matching eval and gt pair) as calculated from gtToEval
        int fn = 0; // False Negatives (segment in gt but no matching eval)
        int fp = 0; // False Positives (segment in eval but no matching gt)

        // number of times an eval segment matches more than one gt segment. This is usually due to the tolerance being
        // set so high that a single eval segment can match multiple gt segments - it can also occur if there are
        // duplicate or overlapping entries in the gt set, though that should not occur.
        int multipleGTMatches = 0;

        // number of times a gt segment matches more than one eval segment. This is usually due to duplicate or
        // overlapping entries in the eval set, though can also occur if the tolerance is set too high.
        int multipleEvMatches = 0;


        // Essentially a merge algorithm: the ..GrpStart and ..GrpEnd variables identify a group of segments in their
        // respective arrays, where a group is a set of segments with the same imageFileName and pageNumber.
        // The outer loop is over compatible groups from the two arrays: Within this loop, we first advance in turn
        // the ..GrpStart vars (always the one that points the segment which is lesser) until they both point to the
        // start of compatible groups. Then we advance the ..GrpEnd vars to point to the first segments after those
        // groups. Finally we iterate over one of the sides setting the gtToEval and evalToGt cross pointers if the
        // segments match.

        int evGrpStart;
        int evGrpEnd = 0;
        int gtGrpStart;
        int gtGrpEnd = 0;

        // advance evGrpStart and gtGrpStart until they refer to segments with the same imageFileName and pageNumber

        while (evGrpEnd < evalArray.length && gtGrpEnd < gtArray.length)
        {
            // Advance ..GrpStart to start the next search for corresponding groups
            evGrpStart = evGrpEnd;
            gtGrpStart = gtGrpEnd;

            int cmp = evalArray[evGrpStart].compareTo(gtArray[gtGrpStart]);
            while (true)
            {
                if (cmp < 0)
                {
                    do
                        ++evGrpStart;
                    while (evGrpStart < evalArray.length &&
                           (cmp = evalArray[evGrpStart].compareTo(gtArray[gtGrpStart])) < 0);
                    if (evGrpStart >= evalArray.length)
                        break;
                }
                else if (cmp > 0)
                {
                    do
                        ++gtGrpStart;
                    while (gtGrpStart < gtArray.length &&
                           (cmp = evalArray[evGrpStart].compareTo(gtArray[gtGrpStart])) > 0);
                    if (gtGrpStart >= gtArray.length)
                        break;
                }
                else
                    break;
            }

            if (evGrpStart >= evalArray.length || gtGrpStart >= gtArray.length)
                break;

            // At this point, ..GrpStart point to the start of compatible segment groups. Now need to fix ..GrpEnd
            // to point to the first segments after the end of the respective groups

            evGrpEnd = evGrpStart + 1;
            while (evGrpEnd < evalArray.length && evalArray[evGrpEnd].compareTo(gtArray[gtGrpStart]) == 0)
                evGrpEnd++;

            gtGrpEnd = gtGrpStart + 1;
            while (gtGrpEnd < gtArray.length && gtArray[gtGrpEnd].compareTo(evalArray[evGrpStart]) == 0)
                gtGrpEnd++;

            // Now the compatible groups are fully identified, we can search for the matching records. If the
            // tolerance is greater than 0, there is no ordering that helps, so we just do a full cartesian product
            // i.e. doubly nested loop.
            for (int gtIndex = gtGrpStart ; gtIndex < gtGrpEnd ; gtIndex++)
            {
                for (int evIndex = evGrpStart ; evIndex < evGrpEnd ; evIndex++)
                {
                    if (gtArray[gtIndex].matchesWithTolerance(evalArray[evIndex], tolerance))
                    {
                        if (evalToGt[evIndex] != -1)
                        {
                            // Don't allow one eval segment to match more than one gt segment (for example, if the
                            // tolerance is too large). We don't do anything clever to choose which matching pair to use
                            // in order to maximise the number of true positives. However, we print a warning if such a
                            // case occurs and record the number of times this occurs
                            System.err.printf("WARNING: At the current tolerance level of %d pixels, evaluation " +
                                              "segment \"%s\" matches at least two ground truth segments: " +
                                              "\"%s\" and \"%s\"%n",
                                              tolerance,
                                              evalArray[evIndex],
                                              gtArray[evalToGt[evIndex]],
                                              gtArray[gtIndex]);
                            multipleGTMatches++;
                        }
                        if (gtToEval[gtIndex] != -1)
                        {
                            // Don't allow one gt segment to match more than one eval segment (for example, if there are
                            // duplicate or overlapping eval segments or the tolerance is too large). We don't do
                            // anything clever to choose which matching pair to use in order to maximise the number of
                            // true positives. However, we print a warning if such a case occurs and record the number
                            // of times this occurs
                            System.err.printf("WARNING: At the current tolerance level of %d pixels, ground truth " +
                                              "segment \"%s\" matches at least two evaluation segments: " +
                                              "\"%s\" and \"%s\"%n",
                                              tolerance,
                                              gtArray[gtIndex],
                                              evalArray[gtToEval[gtIndex]],
                                              evalArray[evIndex]);
                            multipleEvMatches++;
                        }
                        else
                        {
                            gtToEval[gtIndex] = evIndex;
                            evalToGt[evIndex] = gtIndex;
                        }
                    }
                }
            }
        }

        // Any entry in the evalToGt array that is negative corresponds to an evaluation segment for which there
        // is no matching ground truth segment, hence a false positive. If it is non-negative, then that value is the
        // index of the matching ground truth segment, hence a true positive.
        for (int gtIndex : evalToGt)
        {
            if (gtIndex < 0)
                fp++;
            else
                tpEval++;
        }

        for (int evIndex : gtToEval)
        {
            if (evIndex < 0)
                fn++;
            else
                tpGt++;
        }
        // a paranoid sanity check. tpEval must be equal to tpGt so long as the runs were valid, i.e. so long as no
        // multiple matches were found, in which case all bets are off.
        assert (tpEval == tpGt || multipleEvMatches > 0 || multipleGTMatches > 0);
        return (new int[]{tpEval, fp, fn, multipleGTMatches, multipleEvMatches});
    }

    private static void usage()
    {
        System.err.println("ClefSegmentEval: a program to evaluate segmentation clips against a ground truth for");
        System.err.println("    the CLEF-IP Chemical Diagram Segmentation task.");
        System.err.println("Version: " + VERSION);
        System.err.println("Usage: java ClefSegmentEval <ground_truth>.csv <eval_set>.csv [tolerance1 tolerance2 ...]");
        System.err.println("    where:");
        System.err.println("    <ground_truth>.csv is the ground truth csv (comma separated value) file");
        System.err.println("    <eval_set>.csv is the evaluation csv file to evaluate against the ground truth");
        System.err.println(
                "    [tolerance1 tolerance2 ...] is an optional increasing list of non-negative tolerance factors");
        System.err.println("    for matching bounding boxes");
        System.err.println("Input CSV files: The input lines of both the ground truth and the evaluation set CSV");
        System.err.println("    files must contain lines of the form:");
        System.err.println("        US20080096895A1.tif,20,657,2110,563,270");
        System.err.println("    That is, each line describes the bounding box of a molecular diagram segment clipped");
        System.err.println("    from a multipage TIFF format file. Thus each line has 6 comma separated fields where");
        System.err.println("    the first is the name (with no quotes) of the TIFF format file that the segment has");
        System.err.println("    been taken from, the second is the page from that file that the segment was on, and");
        System.err.println("    the remaining fields are, in order, the x, y, width and height coordinates of the");
        System.err.println("    segment.");
        System.err.println("    The input CSV files can optionally have a header line as the first line - detected as");
        System.err.println("    a first line in which the 2nd to 6th fields are not integers. Further any blank");
        System.err.println("    lines or lines for which the first non blank character is a '#' is skipped");
        System.err.println("Output: Output is printed to the standard output stream and is in the form of a CSV file");
        System.err.println("    containing a header row and one row for each tolerance chosen. The fields are:");
        System.err.println("    GroundTruthSize: The number of segments in the ground truth set");
        System.err.println("    EvalSetSize: The number of segments in the evaluation set");
        System.err.println("    Tolerance: The tolerance setting for this evaluation run");
        System.err.println("    TruePositives (TP): The number of true positives - correctly identified segments");
        System.err.println("        found at this tolerance level");
        System.err.println("    FalsePositives (FP): The number of false positives - segments found in the evaluation");
        System.err.println("        set that do not match any segment in the ground truth set at this tolerance level");
        System.err
                .println("    FalseNegatives (FN): The number of false negatives - segments found in the ground truth");
        System.err.println("        set that do not match any segment in the evaluation set at this tolerance level");
        System.err.println("    Precision: TP / (TP + FP): The probability that a segment returned as the bounding");
        System.err.println("        box of a diagram will in fact be so");
        System.err.println("    Recall: TP/ (TP + FN) The probability that the any given real diagram will be");
        System.err.println("        identified and its bounding box returned");
        System.err.println("    F1Measure: Balanced harmonic mean of precision and recall");
        System.err.println("    MultipleGTMatches: The number of times an evaluation segment matches multiple ground");
        System.err.println("        truth segments. This usually happens because the tolerance has been set too high.");
        System.err.println("        Valid evaluation runs should always return 0 for this value");
        System.err.println("    MultipleEvalMatches: The number of times a ground truth segment matches multiple");
        System.err.println("        evaluation segments. This usually happens either because the evaluation set");
        System.err.println("        contains duplicate records, or because the tolerance has been set too high.");
        System.err.println("        Valid evaluation runs should always return 0 for this value");
        System.err.println("Error output: Various error and warning messages are sent to the standard error stream");
        System.err.println("        if their corresponding conditions occur");
        System.err.println("Matching: Segments are considered to match at a tolerance level of T if the image file");
        System.err.println("    and the page number of the segments exactly match and if the bounding boxes");
        System.err.println("    have top, bottom, left and right pixel positions that are with T pixels of each other");
        System.err.println("    respectively");
        System.exit(1);
    }

}

class SegmentBB implements Comparable<SegmentBB>
{
    String imageFileName;
    int pageNumber;
    int x;
    int y;
    int width;
    int height;

    // don't allow very large page numbers or bounding box coordinates as they are almost certainly errors
    private static final int LARGE_INT = 100000;

    private static final int INDEX_SRC_IMAGE = 0;
    private static final int INDEX_PAGE = 1;
//    private static final int INDEX_X = 2;
//    private static final int INDEX_Y = 3;
//    private static final int INDEX_WIDTH = 4;
//    private static final int INDEX_HEIGHT = 5;


    SegmentBB(String imageFileName, int pageNumber, int x, int y, int width, int height)
    {
        this.imageFileName = imageFileName;
        this.pageNumber = pageNumber;
        this.x = x;
        this.y = y;
        this.width = width;
        this.height = height;
    }

    public String toString()
    {
        StringBuilder sb = new StringBuilder(imageFileName);
        sb.append(",");
        sb.append(pageNumber);
        sb.append(",");
        sb.append(x);
        sb.append(",");
        sb.append(y);
        sb.append(",");
        sb.append(width);
        sb.append(",");
        sb.append(height);

        return sb.toString();
    }

    public static SegmentBB[] loadAndSort(String filename, BufferedReader reader)
    {
        ArrayList<SegmentBB> segs = new ArrayList<SegmentBB>(10000);
        String line;
        String trimmedLine;
        int lineno = 1;
        int numbers[] = new int[5];

        try
        {
            for ( ; (line = reader.readLine()) != null ; lineno++)
            {
                trimmedLine = line.trim();
                if (trimmedLine.length() == 0 || trimmedLine.charAt(0) == '#') // empty or comment line
                    continue;

                String fields[] = trimmedLine.split(",");

                if (fields.length != 6)
                {
                    System.err.printf("Error in file \"%s\". Line %d does not have 6 fields: %s%n",
                                      filename,
                                      lineno,
                                      line);
                    continue;
                }
                fields[INDEX_SRC_IMAGE] = fields[INDEX_SRC_IMAGE].trim();
                fields[INDEX_PAGE] = fields[INDEX_PAGE].trim();

                try
                {
                    for (int i = 0 ; i < 5 ; i++)
                    {
                        numbers[i] = Integer.parseInt(fields[i + 1]);
                        if (numbers[i] < 0)
                            throw new NumberFormatException("Integer field values cannot be negative");
                        if (numbers[i] > LARGE_INT)
                            throw new NumberFormatException("Integer field values greater than " +
                                                            LARGE_INT +
                                                            " are assumed to be in error");
                    }
                }
                catch (NumberFormatException e)
                {
                    // If on line 1, assume this is a header and just skip the line. Otherwise print an error message.
                    if (lineno != 1)
                        System.err.printf("Error in number format in file \"%s\", line number %d %s: \"%s\"%n",
                                          filename, lineno, e.getMessage(), line);
                    continue;
                }

                segs.add(new SegmentBB(fields[INDEX_SRC_IMAGE],
                                       numbers[0],
                                       numbers[1],
                                       numbers[2],
                                       numbers[3],
                                       numbers[4]));
            }
        }
        catch (IOException e)
        {
            System.err.printf("Error in reading file \"%s\", line number %d: %s%n", filename, lineno, e.getMessage());
        }
        SegmentBB segArray[] = new SegmentBB[segs.size()];
        segs.toArray(segArray);
        sort(segArray);
        return segArray;
    }


    public int compareTo(SegmentBB that)
    {
        // Note that using subtraction for integer comparison is safe here because we have limited the integers to
        // LARGE_INT. Hence there can be no wraparound.

        // note that we need this to be a total order on groups of segments as defined by imageFileName and pageNumber.
        // i.e. it must show two segments with the same imageFileName and pageNumber to be equal, even (especially)
        // if the bounding boxes are different.
        int val;

        if ((val = this.imageFileName.compareTo(that.imageFileName)) != 0)
            return val;
        return (this.pageNumber - that.pageNumber);
    }

    public boolean matchesWithTolerance(SegmentBB that, int t)
    {
        if (that.x > this.x + t || that.x < this.x - t)
            return false;
        if (that.y > this.y + t || that.y < this.y - t)
            return false;
        if (that.x + that.width > this.x + this.width + t || that.x + that.width < this.x + this.width - t)
            return false;
        if (that.y + that.height > this.y + this.height + t || that.y + that.height < this.y + this.height - t)
            return false;

        return true;

    }
}
